/*
	echoserver: TCP/IP sockets example
	get a connect: keep reading data from the socket, echoing
	back the received data.

	usage:	echoserver [-d] [-p port]
*/

#include <stdio.h>
#include <stdlib.h>	/* needed for os x */
#include <string.h>	/* for memset */
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/errno.h>   /* defines ERESTART, EINTR */
#include <signal.h>      /* for signal() */
#include <sys/wait.h>    /* defines WNOHANG, for wait() */

#include "echo.h"       /* defines default port */

#ifndef ERESTART
#define ERESTART EINTR
#endif

extern int errno;

int debug = 0;
void serve(int port);	/* main server function */
static void sigcld_handler(int);     /* signal handler */

main(int argc, char **argv)
{
	extern char *optarg;
	extern int optind;
	int c, err = 0; 
	int port = SERVICE_PORT;
	static char usage[] = "usage: %s [-d] [-p port]\n";

	while ((c = getopt(argc, argv, "dp:")) != -1)
		switch (c) {
		case 'd':
			debug = 1;
			break;
		case 'p':
			port = atoi(optarg);
			if (port < 1024 || port > 65535) {
				fprintf(stderr, "invalid port number: %s\n", optarg);
				err = 1;
			}
			break;
		case '?':
			err = 1;
			break;
		}
	if (err || (optind < argc)) {
		fprintf(stderr, usage, argv[0]);
		exit(1);
	}
	serve(port);
}

/* serve: set up the service */

void
serve(int port)
{
	int svc;        /* listening socket providing service */
	int rqst;       /* socket accepting the request */
	socklen_t alen;       /* length of address structure */
	struct sockaddr_in my_addr;    /* address of this service */
	struct sockaddr_in client_addr;  /* client's address */
	int sockoptval = 1;
	char hostname[128]; /* host name, for debugging */

	gethostname(hostname, 128);

	/* catch SIGCHLD when a child exits */
	if (signal(SIGCHLD, (void (*)(int))sigcld_handler) == SIG_ERR) {
		perror("server-signal");
		exit(1);
	}

	/* get a tcp/ip socket */
	/*   AF_INET is the Internet address (protocol) family  */
	/*   with SOCK_STREAM we ask for a sequenced, reliable, two-way */
	/*   conenction based on byte streams.  With IP, this means that */
	/*   TCP will be used */

	if ((svc = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		perror("cannot create socket");
		exit(1);
	}

	/* we use setsockopt to set SO_REUSEADDR. This allows us */
	/* to reuse the port immediately as soon as the service exits. */
	/* Some operating systems will not allow immediate reuse */
	/* on the chance that some packets may still be en route */
	/* to the port. */

	setsockopt(svc, SOL_SOCKET, SO_REUSEADDR, &sockoptval, sizeof(int));

	/* set up our address */
	/* htons converts a short integer into the network representation */
	/* htonl converts a long integer into the network representation */
	/* INADDR_ANY is the special IP address 0.0.0.0 which binds the */
	/* transport endpoint to all IP addresses on the machine. */

	memset((char*)&my_addr, 0, sizeof(my_addr));  /* 0 out the structure */
	my_addr.sin_family = AF_INET;   /* address family */
	my_addr.sin_port = htons(port);
	my_addr.sin_addr.s_addr = htonl(INADDR_ANY);

	/* bind to the address to which the service will be offered */
	if (bind(svc, (struct sockaddr *)&my_addr, sizeof(my_addr)) < 0) {
		perror("bind failed");
		exit(1);
	}

	/* set up the socket for listening with a queue length of 5 */
	if (listen(svc, 5) < 0) {
		perror("listen failed");
		exit(1);
	}

	printf("server started on %s, listening on port %d\n", hostname, port);

	/* loop forever - wait for connection requests and perform the service */
	alen = sizeof(client_addr);     /* length of address */

	for (;;) {
		while ((rqst = accept(svc,
		                (struct sockaddr *)&client_addr, &alen)) < 0) {
			/* we may break out of accept if the system call */
			/* was interrupted. In this case, loop back and */
			/* try again */
			if ((errno != ECHILD) && (errno != ERESTART) && (errno != EINTR)) {
				perror("accept failed");
				exit(1);
			}
		}

		/* fork off a process to handle the client request */
		switch (fork()) {
		case -1:
			perror("fork failed\n");
			exit(1);

		default: /* parent */
			close(rqst);
			break;  /* continue looping for next connection */

		case 0:  /* child */
			close(svc);  /* child doesn't need this */
			dostuff(rqst); /* run the service */
			exit(0);
		}
	}
}

/* signal handler - note when a child process dies */

static void
sigcld_handler(int signum)
{
	int status;
	pid_t pid;

#ifdef LINUX
	signal(SIGCHLD, (void (*)(int))sigcld_handler);
#endif

	while ((pid = waitpid(-1, &status, WNOHANG)) > 0) {
		printf("server child %d exited with return value %d\n",
		        pid, WEXITSTATUS(status));
	}
}

/* dostuff: process client requests */

dostuff(int fd) {
	int pid = getpid();
	int len;
	char data[256];

	if (debug)
		printf("spawned process %d: \n", pid);
	while (getline(data, 256, fd)) {
		if (strcmp(data, ".") == 0)
			break;
		len = strlen(data);
		if (debug)
			printf("process %d: received \"%s\"\n", pid, data);
		fflush(stdout);
		/* echo it back to the sender */
		data[len] = '\n';
		write(fd, data, len+1);
	}
	if (debug) printf("process %d: exiting\n", pid);
}
